from datetime import datetime, timedelta, timezone
import json
from lclazy import LazyLoader
from lcconfig import generate_totp
try:
    from lchttp import json_dumps
except (ImportError, ModuleNotFoundError):
    from json import dumps as json_dumps
import time
import logging
import secrets
from typing import Literal, TYPE_CHECKING
from .typehints import CID_STR, PasetoFooter, PasetoPayload, Subject

if TYPE_CHECKING:
    from lcrequests.keys import FacadePrivateKey, FacadePublicKey
    import jwt
    import pyseto
else:
    jwt = LazyLoader('jwt', globals(), 'jwt')
    pyseto = LazyLoader('pyseto', globals(), 'pyseto')

logger = logging.getLogger(__name__)
TIME_LEEWAY = 20


class Token:
    """
    Token class for generating signed JWTs. Pass to Request
    class so that each request gets its own signed token.
    """

    def __init__(
        self,
        private_key: 'FacadePrivateKey',
        auth_method: Literal["SIGNED_JWT", "AUTO_JWT"] = 'SIGNED_JWT',
        cid: str = '',
        email: str = '',
        public_key_cid: str = '',
        sub: str = '',
    ):
        self.private_key = private_key
        self.auth_method = auth_method
        self.cid = cid
        self.email = email
        self.public_key_cid = public_key_cid
        self.public_key = private_key.public_key
        self.private_key_pem = private_key.as_pem
        self.sub = sub

    def __repr__(self) -> str:
        name = self.__class__.__name__
        return f"{name}('<private_key>', {self.auth_method!r}, " \
               f"{self.cid!r}, {self.email!r}, {self.public_key_cid!r})"

    def __str__(self) -> str:
        return f"{self.auth_method} <token>"

    def auth_header(self, include_otp=False, digits=8) -> str:
        """
        Create an HTTP Authorization header

        Authorization: JWT 401f7ac837da42b97f613d789819ff93537bee6a

        :param include_otp: Include an OTP value in the JSON token data for additional verification.
        :param digits: Length of the OTP.
        """
        if include_otp:
            otp = generate_totp(self.public_key.as_pem.decode(),
                                digits=digits).now()
        else:
            otp = ''

        token = self.sign(otp=otp)
        return f"{self.auth_method} {token}"

    def sign(self, otp: str = '') -> str:
        """
        Create and return signed authentication JWT
        """
        token_data = {
            "time": int(time.time()),
            "nonce": secrets.token_urlsafe(nbytes=8),
        }
        if self.sub:
            token_data['sub'] = self.sub
        if self.public_key_cid:
            token_data['public_key_cid'] = self.public_key_cid
        if self.cid:
            token_data['cid'] = str(self.cid)
        elif self.email:
            token_data['email'] = self.email
        if otp:
            token_data['otp'] = otp

        headers = {
            "kid": self.private_key.public_key.fingerprint,
        }
        token = jwt.encode(
            payload=token_data,
            key=self.private_key_pem,
            algorithm=self.public_key.allowed_algorithms[0],
            headers=headers,
        )
        return token


class RegistrationToken(Token):
    """
    Use this token on the client side when registering
    new accounts / public keys on the server.
    """

    def __init__(
        self,
        private_key: 'FacadePrivateKey',
        auth_method: Literal["SIGNED_JWT", "AUTO_JWT"] = 'AUTO_JWT',
        cid: str = '',
        email: str = '',
    ):
        super().__init__(
            private_key=private_key,
            auth_method=auth_method,
            cid=cid,
            email=email,
        )

    def auth_header(self, include_otp=True, digits=8) -> str:
        return super().auth_header(include_otp=include_otp, digits=digits)


# ---------------------------------------------------------------------------
class PasetoToken:
    """
    Represents a Paseto Token that's either been constructed
    by our code or has been verified to be valid.
    """
    auth_method = 'Paseto'
    version: int = 4
    subject: Subject = ''

    def __init__(
        self,
        private_key: 'FacadePrivateKey',
        user_cid: CID_STR = '',
        email: str = '',
        subject_id: CID_STR = '',
        verification_cid: CID_STR = '',
        expires: int = 90,
    ):
        self.private_key: 'FacadePrivateKey' = private_key
        self.public_key: 'FacadePublicKey' = private_key.public_key
        self.user_cid = user_cid
        self.email: str = email
        self.subject_id: CID_STR = subject_id
        self.verification_cid: CID_STR = verification_cid
        self.expires = expires

    def __repr__(self) -> str:
        name = self.__class__.__name__
        return f"{name}('<private_key>', {self.auth_method!r}, " \
               f"{self.user_cid!r}, {self.email!r}, {self.expires!r})"

    def __str__(self) -> str:
        return f"{self.auth_method} <token>"

    def auth_header(self) -> str:
        """
        Create an HTTP Authorization header.
        """
        return f"{self.auth_method} {self.sign().decode()}"

    def payload(self) -> PasetoPayload:
        """
        Return data for Paseto token payload.
        """
        now = datetime.now(tz=timezone.utc)
        payload = PasetoPayload(
            nonce=secrets.token_urlsafe(nbytes=8),
            exp=(now + timedelta(seconds=self.expires)).isoformat(
                timespec="seconds"),
            nbf=(now -
                 timedelta(seconds=TIME_LEEWAY)).isoformat(timespec="seconds"),
        )

        if self.user_cid:
            payload['user_cid'] = self.user_cid
        elif self.email:
            payload['email'] = self.email

        return payload

    def footer(self) -> PasetoFooter:
        """
        Insert keys into the footer to enable server-side
        verification, such as lookup of Public Keys to verify
        token signing.
        """
        if self.verification_cid:
            footer = PasetoFooter(cid=self.verification_cid)
        else:
            kid = self.private_key.public_key.as_paseto(
                version=self.version).to_paserk_id()
            footer = PasetoFooter(kid=kid)

        if self.subject:
            footer['sub'] = self.subject
        if self.subject_id:
            footer['sub_id'] = self.subject_id

        return footer

    def sign(self) -> bytes:
        """
        Create and return signed authentication Paseto token
        """
        paseto = pyseto.Paseto.new(
            exp=self.expires,
            include_iat=True,
            leeway=TIME_LEEWAY,
        )

        return paseto.encode(
            key=self.private_key.as_paseto(self.version),
            payload=json_dumps(self.payload()),
            footer=json_dumps(self.footer()),
            serializer=json,
        )


# ---------------------------------------------------------------------------
class V1PasetoToken(PasetoToken):
    version = 1


class V2PasetoToken(PasetoToken):
    version = 2


class V3PasetoToken(PasetoToken):
    version = 3


V4PasetoToken = PasetoToken


# ---------------------------------------------------------------------------
class RegistrationPasetoToken(PasetoToken):
    """
    This token must be used when auto-registering new EncryptionKeys.
    """
    auth_method = 'Auto_Paseto'

    def footer(self) -> PasetoFooter:
        """
        Add a one-time password value to provide some extra
        security against malicious registration attempts.
        """
        data = super().footer()
        data['otp'] = generate_totp(
            self.private_key.public_key.as_pem.decode(), digits=8).now()
        if not data.get('sub'):
            raise pyseto.PysetoError(
                'Registration token must contain "sub" claim')
        return data

    def payload(self) -> PasetoPayload:
        data = super().payload()
        data['aud'] = 'register'
        return data


# ---------------------------------------------------------------------------
class V1RegistrationPasetoToken(RegistrationPasetoToken):
    version = 1


class V2RegistrationPasetoToken(RegistrationPasetoToken):
    version = 2


class V3RegistrationPasetoToken(RegistrationPasetoToken):
    version = 3


V4RegistrationPasetoToken = RegistrationPasetoToken

__all__ = (
    'Token',
    'RegistrationToken',
    'PasetoToken',
    'V1PasetoToken',
    'V2PasetoToken',
    'V3PasetoToken',
    'V4PasetoToken',
    'RegistrationPasetoToken',
    'V2RegistrationPasetoToken',
    'V1RegistrationPasetoToken',
    'V3RegistrationPasetoToken',
    'V4RegistrationPasetoToken',
)
